# -*- coding: utf-8 -*-

from __future__ import generators,absolute_import

"""\
This class is a Twisted-compatible frontend to sqlmix.Db.

It has the same interface, except that all Do* methods return a Deferred.
Internally, it works by delegating all SQL commands to a separate thread.

 >> dbi = sqlmix.DbPool([args of sqlmix.Db])
 >>
 >> @inlineCallbacks
 >> def foo(what_id):
 >>     with dbi() as db:
 >>         yield db.Do("delete from whatever where id=${id}", id=what_id, _empty=1)
 >>         res = yield db.DoSelect("select id from whatever", _empty=1)
 >>         for x in res:
 >>            print res
 >>         yield db.commit()
"""
#
#    Copyright (C) 2011 Matthias urlichs <smurf@smurf.noris.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sqlmix
from time import time,sleep
import string
import re
import sys
from traceback import print_exc
from zope.interface import implements
from twisted.application import service
from twisted.internet import reactor
from twisted.internet.defer import Deferred,DeferredList,maybeDeferred,inlineCallbacks,returnValue,succeed
from twisted.python import log
from twisted.python.failure import Failure
from twisted.python.threadpool import ThreadPool
from twisted.internet import threads
from threading import Lock
from Queue import Queue

__all__ = ('DbPool','NoData','ManyData')

_DEBUG = False

NoData = sqlmix.NoData
ManyData = sqlmix.ManyData

def _call(r,p,a,k):
	"""Drop the first argument (i.e. lose the Deferred result)"""
	return p(*a,**k)

def tname():
	import threading
	try:
		return threading._active[threading._get_ident()].name
	except KeyError:
		return "???"

def _print_error(f):
	f.printTraceback(file=sys.stderr)

def debug_(flag,*a):
	if not flag: return
	def pr(x):
		if isinstance(x,(tuple,list)):
			return "\n".join((pr(y).strip("\n") for y in x))
		elif isinstance(x,(basestring,int)):
			try:
				return unicode(x)
			except:
				return repr(x)
		else:
			return repr(x)
	s=" ".join((pr(x) for x in a))

	if flag is True: fl=""
	else: fl=flag+": "
	sys.stderr.write(fl+s+"\n")
	sys.stderr.flush()
def debug(*a):
	debug_(_DEBUG,*a)

class DbPool(object,service.Service):
	"""\
	Manage a pool of database connections.

	TODO: shrink the pool.
	TODO: issue periodic keepalive requests.
	"""
	timeout = 70 # one minute plus
	implements(service.IService)

	def __init__(self,*a,**k):
		"""\
		Create a pool of database connections, for processing (a sequence of)
		SQL commands in the background.
		"""
		k['_single_thread'] = True

		self.db = []
		self.args = a
		self.kwargs = k
		self.lock = Lock()
		self.cleaner = None
		self._tb = {}
		self.stopping = False
		self.threads = ThreadPool(minthreads=2, maxthreads=100, name="Database")
		self.threads.start()
		#reactor.addSystemEventTrigger('before', 'shutdown', self.stop)
		reactor.addSystemEventTrigger('after', 'shutdown', self._dump)
		reactor.addSystemEventTrigger('after', 'shutdown', self.stop2)

	def stop2(self):
		if self.db is not None:
			for db in self.db:
				db[0].close("AfterShutdown Service")
		self.threads.stop()

	def stop(self):
		self.stopping = True

	def _get_db(self,tid=None):
		if self.db:
			r = self.db.pop()[0]
			s="OLD"
		else:
			r = _DbThread(self)
			s="NEW"

		if tid:
			debug(s, r.tid,tid)
			r.tid=tid
		else:
			debug(s, r.tid)

		return r

	def _put_db(self,db):
		if self.db is None or self.stopping:
			db.close("Shutdown")
			return db.done
		if db.q is None:
			raise RuntimeError("Queueing closed DB handle")
		for d in self.db:
			if db is d[0]:
				raise RuntimeError("DoubleQueued")
		db.count = 0
		try:
			t = time()+self.timeout
			self.db.append((db,t))
			if self.cleaner is None:
				self.cleaner = reactor.callLater(self.timeout,self._clean)
		except Exception:
			print_exc()
		else:
			debug("BACK",db.tid)
	
	def _clean(self):
		self.cleaner = None
		t = time()
		while self.db and self.db[0][1] <= t:
			db = self.db.pop(0)[0]
			db.close("Timeout")
		if self.db:
			self.cleaner = reactor.callLater(self.db[0][1]-t,self._clean)
	def __del__(self):
		if self.cleaner:
			reactor.cancelCallLater(self.cleaner)
			self.cleaner = None
		while self.db:
			db = self.db.pop(0)[0]
			db.close("Nonref Parent")

	def stopService(self):
		super(DbPool,self).stopService()
		if self.cleaner:
			self.cleaner.cancel()
			self.cleaner = None
		dbl = self.db
		self.db = None
		dl = []
		for db in dbl:
			db = db[0]
			db.close("Shutdown Service")
			dl.append(db.done)
		return DeferredList(dl)
		
	def __call__(self, job=None,retry=0):
		"""\
		Get a new connection from the database pool (or start a new one)
		and return a thread handler.

		Usage:
		>>> @inlineCallbacks
		>>> def something(...):
		>>> 	dbpool = DbPool(...) # arguments like sqlmix.Db()
		>>>		with dbpool() as db:
		>>>			d = db.Do("...")
		>>>			assert(isinstance(d,twisted.internet.defer.Deferred))
		>>>			res = yield d

		The transaction will be committed if you leave the "with" block
		normally or with a CommitThread exception. Otherwise, it will be
		rolled back.

		Note that you must use the @inlineCallbacks method if you want
		to use the database connection within the block. Otherwise, control
		will have left the "with" block and the connection will be dead.

		Alternately, you can pass a procedure and an optional repeat count:
		>>> def proc(db):
		>>>     d = db.Do("...")
		>>>     return d
		>>> d = dbpool(proc, 10)

		The procedure will be retried up to 10 times if there are errors;
		if they persist, the first error will be re-raised.

		"""
		if not job:
			return self._get_db()
		return self._call(job,retry)

	@inlineCallbacks
	def _call(self, job, retry):
		global tid
		tid += 1
		mtid = tid
		debug("STARTCALL",job,retry,mtid)

		e1 = None
		try:
			while True:
				db = self._get_db(mtid)
				self._note(db)
				try:
					debug("CALL JOB",mtid)
					d = job(db)
					debug("RET JOB",mtid,d)
					def pr(r):
						debug("RES JOB",mtid,r)
						return r
					d.addBoth(pr)
					res = yield d
				except (EnvironmentError,NameError):
					self._denote(db)
					yield db.rollback()
					raise
				except Exception:
					self._denote(db)
					yield db.rollback()
					if retry:
						retry -= 1
						continue
					raise
				except BaseException:
					self._denote(db)
					yield db.rollback()
					raise
				else:
					self._denote(db)
					if isinstance(res,BaseException):
						yield db.rollback()
						if isinstance(res,(EnvironmentError,NameError)):
							returnValue( res )
						elif isinstance(res,Exception):
							if retry:
								retry -= 1
								continue
							returnValue( res )
						else: # BaseException
							returnValue( res )
					else:
						yield db.commit()
					returnValue( res )
		finally:
			debug("ENDCALL",job,retry)

	def _note(self,x):
		if not _DEBUG: return
		import inspect
		self._tb[x.tid] = inspect.stack(1)
	def _denote(self,x):
		if not _DEBUG: return
		del self._tb[x.tid]
	def _dump(self):
		if not _DEBUG: return
		for a,b in self._tb.items():
			#(<frame object at 0x8a1b724>, '/mnt/daten/src/git/sqlmix/sqlmix/twisted.py', 250, '_note', ['\t\tself._tb[x.tid] = inspect.stack(1)\n'], 0)

			print >>sys.stderr,"Stack",a
			for fr,f,l,fn,lin,lini in b[::-1]:
				if fn == "__call__": break
				print >>sys.stderr,"Line %d in %s: %s" % (l,f,fn)
				print >>sys.stderr,"\t"+lin[lini].strip()

def _do_callback(tid,d,res):
	debug("DO_CB",tid,d,res)
	d.callback(res)
	debug("DID_CB",tid,d,res)

tid = 0
class _DbThread(object):
	def __init__(self,parent):
		global tid
		tid += 1
		self.tid = tid
		self.parent = parent
		self.q = Queue()
		debug("INIT",self.tid)
		self.done = threads.deferToThreadPool(reactor, self.parent.threads, self.run,self.q)
		self.started = False
		self.count = 0

		self.committed = []
		self.rolledback = []

	def __repr__(self):
		return "<_DbT.%d>"%(self.tid,)

	def __enter__(self):
		self.parent._note(self)
		return self
	def __exit__(self, a,b,c):
		self.parent._denote(self)
		if self.q is None:
			return False
		if b is None or isinstance(b,sqlmix.CommitThread):
			d = self.commit()
		else:
			from traceback import format_exception
			debug("EXIT ON ERROR",format_exception(a,b,c))
			d = self.rollback()
		d.addErrback(log.err)
		return False

	def call_committed(self,proc,*a,**k):
		self.committed.append((proc,a,k))
	def call_rolledback(self,proc,*a,**k):
		self.rolledback.append((proc,a,k))

	def _run_committed(self,r):
		return self._run_(r,self.committed,"COMMIT")
	def _run_rolledback(self,r):
		return self._run_(r,self.rolledback,"ROLLBACK")
	def _run_(self,r,rc,rcname):
		self.committed = []
		self.rolledback = []
		d = succeed(None)
		for proc,a,k in rc[::-1]:
			debug("AFTER "+rcname,proc,a,k)
			d.addCallback(_call,proc,a,k)
			d.addErrback(log.err)
		d.addCallback(lambda _: r)
		return d


	def run(self,q):
		try:
			db = sqlmix.Db(*self.parent.args,**self.parent.kwargs)
		except BaseException:
			"""No go. Return that error on every call."""
			f = Failure()
			while True:
				d,proc,a,k = q.get()
				if not d: break
				reactor.callFromThread(d.errback,f)
			return
		debug("START",self.tid, tname())
		res = None
		d = True
		while d:
			d = None
			sent = False
			try:
				d,proc,a,k = q.get()
				res = None
				debug("DO",self.tid,d,proc,a,k)
				if proc == "COMMIT":
					db.commit()
					res = k.get('res',None)
				elif proc == "ROLLBACK":
					db.rollback()
					res = k.get('res',None)
				else:
					r = getattr(db,proc)
					debug("CALL",self.tid,r)
					res = r(*a,**k)
					debug("CALLED",self.tid,res)
			except BaseException as e:
				res = Failure()
				if d:
					debug("EB",self.tid,d,res)
					reactor.callFromThread(d.errback,res)
					sent = True
				else:
					debug("ERR",self.tid,res)
					raise
			finally:
				if d and not sent:
					debug("CB",self.tid,d,res)
					reactor.callFromThread(_do_callback,self.tid,d,res)
				debug("DID",self.tid,proc)
		db.close()
		debug("STOP",self.tid)
		return

	def close(self,reason="???"):
		if self.q is None:
			if reason != "__del__":
				debug("DEAD_CALLED_TWICE",self.tid,reason)
			return
		self.q.put((None,"ROLLBACK",[],{}))
		debug("DEAD",self.tid,reason)
		self.q = None

	def __del__(self):
		self.close("__del__")

	def commit(self,res=None):
		d = Deferred()
		if self.count:
			debug("CALL COMMIT",self.tid,d,res)
			self.q.put((d,"COMMIT",[],{'res':res}))
		else:
			debug("NO CALL COMMIT",self.tid,d,res)
			d.callback(res)
		d.addCallback(self._run_committed)
		d.addErrback(self._run_rolledback)
		d.addBoth(self._done)
		return d

	def rollback(self,res=None):
		d = Deferred()
		if self.q is None:
			d.callback(res)
			return d
		if self.count:
			debug("CALL ROLLBACK",self.tid,d,res)
			self.q.put((d,"ROLLBACK",[],{'res':res}))
		else:
			debug("NO CALL ROLLBACK",self.tid,d,res)
			d.callback(res)
		d.addBoth(self._run_rolledback)
		d.addBoth(self._done)
		return d

	def release(self,d):
		"""Convenience method: add myself to a Deferred to stop processing"""
		d.addCallbacks(self.commit,self.rollback)

	def _done(self,r):
		d = maybeDeferred(self.parent._put_db,self)
		d.addCallbacks(lambda _: r, lambda _: _)
		return d

	def _do(self,job,*a,**k):
		"""Wrapper for calling the background thread."""
		self.count += 1
		debug = k.get("_debug",_DEBUG)
		d = Deferred()
		debug_(debug,"QUEUE",self.tid,job,a,k)
		self.q.put((d,job,a,k))
		def _log(r):
			debug_(debug,"DEQUEUE",self.tid,r)
			return r
		d.addBoth(_log)
		return d
		
	def Do(self,*a,**k):
		return self._do("Do",*a,**k)
	def DoFn(self,*a,**k):
		return self._do("DoFn",*a,**k)
	def DoSelect(self,*a,**k):
		k["_store"] = 1
		return self._do("DoSelect",*a,**k)
	Do.__doc__ = sqlmix.Db.Do.__doc__ + "\nReturns a Deferred.\n"
	DoFn.__doc__ = sqlmix.Db.DoFn.__doc__ + "\nReturns a Deferred.\n"
	DoSelect.__doc__ = sqlmix.Db.DoSelect.__doc__ + "\nReturns a Deferred.\n"




# -*- coding: utf-8 -*-

from __future__ import generators,absolute_import

"""\
This module provides a thread-safe way to access any Python SQL database with
* A common and clear syntax for placing values in SQL statements.
   >> foo,bar = db.DoFn("select ... where id=${key}", key=whatever)
* single-line SQL commands
* Common parameters for database setup (no more "db" vs. "database")
* Single-line SQL statements and single-row SELECTs:
   >> db.Do("insert ...")
   >> foo,bar = db.DoFn("select ... where id=123")
* SELECT statements as iterators:
   >> for a,b in db.DoSelect("select ..."):
   >>     ...
"""
#
#    Copyright (C) 2011 Matthias urlichs <smurf@smurf.noris.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from time import time,sleep
import sys
import os
import re
from sys import exc_info
from threading import local

class CommitThread(Exception):
	u"""\
		If you leave a database handler's with … block by raising an
		exception descending from this class, the transaction will be
		committed instead of being rolled back.
		"""
	pass

class FakeLocal(object):
	"""\
		This connection is only used by one thread,
		so don't bother with thread-local variables.
		"""
	pass

__all__ = ["Db","NoData","ManyData"]

def fixup_error(cmd):
	"""Append the full command to the error message"""
	e1,e2,e3 = exc_info()
	k=list(e2.args)
	k.append(cmd)
	e2.args=tuple(k)

class _NOTGIVEN: pass

class db_data(object):
	sequential = False
	_store = 1 # safe default
	_cursor = True
	def __init__(self, **kwargs):
		"""standard keywords: host,port,database,username,password"""
		for f in "host port database username password".split():
			v = kwargs.pop(f,_NOTGIVEN)
			if v is _NOTGIVEN:
				continue
			if f == "port":
				v=int(v)
			setattr(self,f,v)
		kwargs.setdefault("charset","utf8")
		self.kwargs = kwargs

class _db_mysql(db_data):
	_store = 1
	host="localhost"
	port=3306
	def __init__(self, **kwargs):
		self.DB = __import__("MySQLdb")
		self.DB.cursors = __import__("MySQLdb.cursors").cursors
		super(_db_mysql,self).__init__(**kwargs)

	def conn(self):
		return self.DB.connect(db=self.database, host=self.host, user=self.username, passwd=self.password, port=self.port, **self.kwargs)

class _db_ultramysql(db_data):
	_store = 1
	_cursor = False
	host="localhost"
	port=3306
	def __init__(self, **kwargs):
		self.DB = __import__("umysql")
		super(_db_ultramysql,self).__init__(**kwargs)
		self.DB.paramstyle = 'format'

	def conn(self):
		c = self.DB.Connection()
		c.connect(self.host, self.port, self.username, self.password, self.database, False, "utf8")
		return c

class _db_odbc(db_data):
	def __init__(self, **kwargs):
		self.DB = __import__("mx.ODBC.iODBC")
		super(_db_odbc,self).__init__(**kwargs)

	def conn(self):
		if self.port:
			self.host = self.host+":"+str(self.port)
			self.port=None

		return self.DB.connect(database=self.database, host=self.host, user=self.username, password=self.password)

class _db_postgres(db_data):
	def __init__(self, **kwargs):
		self.DB = __import__("psycopg2")
		super(_db_postgres,self).__init__(**kwargs)

	def conn(self):
		return self.DB.connect(database=self.database,host=self.host, user=self.username, password=self.password, port=self.port)

class _db_sqlite(db_data):
	sequential = True
	def __init__(self, **kwargs):
		self.DB = __import__("sqlite3.dbapi2")
		if hasattr(self.DB,"dbapi2"): self.DB=self.DB.dbapi2
		super(_db_sqlite,self).__init__(**kwargs)

	def conn(self):
		return self.DB.connect(self.database)

_databases = {
	    "mysql": _db_mysql,
	    "odbc": _db_odbc,
	    "postgres": _db_postgres,
	    "sqlite": _db_sqlite,
	    "ultramysql": _db_ultramysql,
}


## possible parsers

## qmark: '?'
def _init_qmark():
	return []
def _do_qmark(name, arg, params):
	arg.append(params[name])
	return "?"
def _done_qmark(cmd,args):
	return (cmd,tuple(args))

## numeric: ':n'
def _init_numeric():
	return []
def _do_numeric(name, arg, params):
	arg.append(params[name])
	return ":" + str(len(arg))
def _done_numeric(cmd,args):
	return (cmd,args)

## named: ':name'

def _init_named():
	return {}
def _do_named(name, arg, params):
	arg[name] = params[name]
	return ":" + name
def _done_named(cmd,args):
	return (cmd,args)

## format: '%s'

def _init_format():
	return []
def _do_format(name, arg, params):
	arg.append(params[name])
	return "%s"
def _done_format(cmd,args):
	return (cmd,args)

## pyformat: '%(name)s'

def _init_pyformat():
	return {}
def _do_pyformat(name, arg, params):
	arg[name] = params[name]
	return "%(" + name + ")s"
def _done_pyformat(cmd,args):
	return (cmd,args)

_parsers = {
		"qmark"   : (_init_qmark,_do_qmark,_done_qmark),
		"numeric" : (_init_numeric,_do_numeric,_done_numeric),
		"named"   : (_init_named,_do_named,_done_named),
		"format"  : (_init_format,_do_format,_done_format),
		"pyformat": (_init_pyformat,_do_pyformat,_done_pyformat),
 	   }


## these might be based on StopIteration instead.  Too dangerous ???

class NoData(Exception):
	pass
class ManyData(Exception):
	pass
#class NoDatabase(Exception):
#	pass

class DbPrep(object):
	"""Base class for command prep"""
	def __init__(self):
		if hasattr(self.DB,'paramstyle'):
			paramstyle = self.DB.paramstyle
		else:
			paramstyle = self.DB.DB.paramstyle
		(self.arg_init, self.arg_do, self.arg_done) \
			 = _parsers[paramstyle]
	def prep(self,_cmd,**kwargs):

		args = self.arg_init()
		def _prep(name):
			return self.arg_do(name.group(1), args, kwargs)
		
		_cmd = re.sub(r"\$\{([a-zA-Z][a-zA-Z_0-9]*)\}",_prep,_cmd)
		return self.arg_done(_cmd,args)
		

class Db(DbPrep):
	"""\
	Main database connection object.

	Internally, manages one back-end connection per thread.

	Possible keywords: dbtype,host,port,database,username,password; config=inifile,cfg=section
	"""

	# These variables cache whether the database supports turning off
	# autocommit, or setting the read wait timeout.
	_set_ac1 = True
	_set_ac2 = True
	_set_timeout = True
	_set_isolation = True

	def __init__(self, cfg=None, **kwargs):
		if cfg is not None:
			try:
				cffile = kwargs.pop('config')
			except KeyError:
				from os.path import expanduser as home
				cffile = home("~/.sqlmix.conf")
			if isinstance(cffile,str):
				from configparser import ConfigParser
				cfp = ConfigParser()
				cfp.read(cffile)
			else:
				cfp = cffile
			args = dict(cfp.items(cfg))
			args.update(kwargs)
			kwargs = args

		self._trace = kwargs.pop("trace",None)

		dbtype = kwargs.pop("dbtype","mysql")
		self.DB = _databases[dbtype](**kwargs)
		self.DB.dbtype=dbtype
		if self._trace is not None:
			self._trace("INIT",dbtype,kwargs)

		if kwargs.pop("_single_thread",False):
			self._c = FakeLocal()
		else:
			self._c = local()
		self._c.conn=None

#		if dbtype == "mysql":
#			self.CArgs = (self.DB.DB.cursors.CursorNW,)
#		else:
		self.CArgs = ()
		self.isolation = kwargs.pop("isolation",None)
#

		super(Db,self).__init__()
		
	def _conn(self, skip=False):
		"""Create a connection to the underlying database."""
		
		if not hasattr(self._c,"conn") or self._c.conn is None:
			if skip: return None

			Ex=None
			r = self.DB.conn()

			if self._set_ac1:
				try: r.setconnectoption(self.DB.DB.SQL.AUTOCOMMIT, self.DB.DB.SQL.AUTOCOMMIT_OFF)
				except AttributeError: self._set_ac1 = False
				else: self._set_ac2 = False
			if self._set_ac2:
				try:
					if self.DB._cursor:
						r.cursor(*self.CArgs).execute("SET AUTOCOMMIT=0")
					else:
						r.query("SET AUTOCOMMIT=0")
				except Exception: self._set_ac2 = False
		
			if self._set_timeout:
				try: 
					if self.DB._cursor:
						r.cursor(*self.CArgs).execute("SET WAIT_TIMEOUT=7200") # 2h
					else:
						r.query("SET WAIT_TIMEOUT=7200")
				except Exception: self._set_timeout = False

			if self._set_isolation and self.isolation:
				try:
					if self.DB._cursor:
						r.cursor(*self.CArgs).execute("SET SESSION TRANSACTION ISOLATION LEVEL "+self.isolation)
					else:
						r.query("SET SESSION TRANSACTION ISOLATION LEVEL "+self.isolation)
						
				except Exception: self._set_isolation = False

#			try:
#				r.stringformat = self.DB.DB.UNICODE_STRINGFORMAT
#				r.encoding = 'utf-8'
#			except AttributeError:
#				pass

			self._commit(r)
			self._c.conn = r
		#r.cursor(*self.CArgs).execute('BEGIN')
		return self._c.conn

	def _commit(self,c,cmd="commit"):
		if(hasattr(c,cmd)):
			getattr(c,cmd)()
		elif self.DB._cursor:
			c.cursor(*self.CArgs).execute(cmd)
		else:
			c.query(cmd)

	def close(self):
		"""\
		Close the current connection, if any.
		"""
		self.rollback()

		c = self._conn(skip=True)
		if c:
			if self._trace is not None:
				self._trace("Close","","")
			c.close()
		else:
			if self._trace is not None:
				self._trace("Close","NoConn","")

	def commit(self):
		"""\
		Commit the current transaction.
		
		This calls all procedures that have been registered with `call_committed`
		in reverse order, and discards all calls registered with `call_rolledback`.
		"""
		if self._trace is not None:
			self._trace("Commit","","")
		c = self._conn(skip=True)
		if c:
			self._commit(c)

		# callbacks
		self._c.rolledback = None
		if hasattr(self._c,"committed") and self._c.committed:
			x = self._c.committed
			self._c.committed = None
			for proc,a,k in x[::-1]:
				proc(*a,**k)

		#self._conn.cursor(*self.CArgs).execute('BEGIN')

	def rollback(self):
		"""\
		Roll back the current transaction.
		
		This calls all procedures that have been registered with `call_rolledback`
		in reverse order, and discards all calls registered with `call_committed`.
		"""
		if self._trace is not None:
			self._trace("RollBack","","")
		c = self._conn(skip=True)
		if c:
			self._commit(c,"rollback")

		# cancel callbacks
		self._c.committed = None
		if hasattr(self._c,"rolledback") and self._c.rolledback:
			x = self._c.rolledback
			self._c.rolledback = None
			for proc,a,k in x[::-1]:
				proc(*a,**k)

		#self._conn.cursor(*self.CArgs).execute('BEGIN')

	def call_committed(self,proc,*a,**k):
		"""\
		Remember to call a procedure (with given parameters) after committing the
		current transaction.
		"""
		if not hasattr(self._c,"committed") or self._c.committed is None:
			self._c.committed = []
		self._c.committed.append((proc,a,k))

	def call_rolledback(self,proc,*a,**k):
		"""\
		Remember to call a procedure (with given parameters) after rolling back the
		current transaction.
		"""
		if not hasattr(self._c,"rolledback") or self._c.rolledback is None:
			self._c.rolledback = []
		self._c.rolledback.append((proc,a,k))

	def DoFn(self, _cmd, **kv):
		"""Select one (and only one) row.

		Example:
		>>>	i,before = db.DoFn("select id,last_login from sometable where name=${myname}", myname="Fred")

		This function always returns an array of values.
		Thus, single-value assignments need a comma after the variable name.
		>>>	i, = db.DoFn("select id from sometable where name=${myname}", myname="Fred")
		
		Special keywords:

		_dict is True: return a column/value dictionary instead of an array.
		_dict is a type: as before, but use that. 
		>>>	info = db.DoFn("select * from sometable where name=${myname}",myname="Fred", _dict=True)

		"""
		conn=self._conn()
		_cmd = self.prep(_cmd, **kv)
		try:
			if self.DB._cursor:
				curs=conn.cursor(*self.CArgs)
				curs.execute(*_cmd)
			else:
				curs=conn.query(*_cmd)
		except:
			fixup_error(_cmd)
			raise

		if hasattr(curs,'fetchone'):
			val = curs.fetchone()
		elif not curs.rows:
			val = None
		else:
			val = curs.rows.pop(0)

		if self._trace is not None:
			self._trace("DoFn",_cmd,val)
		if not val:
			raise NoData(_cmd)

		as_dict=kv.get("_dict",None)
		if as_dict:
			if as_dict is True:
				as_dict = dict
			names = map(lambda x:x[0], curs.description)

		if curs.fetchone() if hasattr(curs,'fetchone') else curs.rows:
			raise ManyData(_cmd)

		if as_dict:
			val = as_dict(zip(names,val))
		return val

	def Do(self, _cmd, **kv):
		"""Database-specific Do function"""
		conn=self._conn()
		_cmd = self.prep(_cmd, **kv)

		try:
			if self.DB._cursor:
				curs=conn.cursor(*self.CArgs)
				curs.execute(*_cmd)
			else:
				curs=conn.query(*_cmd)
		except:
			fixup_error(_cmd)
			raise

		if isinstance(curs,(tuple,list)): # ultramysql
			r = curs[1]
			if not r:
				r = curs[0]
		else:
			r = curs.lastrowid
			if not r:
				r = curs.rowcount

		if self._trace is not None:
			self._trace("DoFn",_cmd,r)
		if r == 0 and not '_empty' in kv:
			raise NoData(_cmd)
		return r

	def DoSelect(self, _cmd, **kv):
		"""Select one or more rows from a database.

		>>>	for i,name in db.DoSelect("select id,name from sometable where name like ${pattern}", pattern='f%d'):
		...		print i,name
		
		The iterator always returns an array of values.
		Thus, single-value assignments need a comma after the variable name.
		>>>	for name, in db.DoSelect("select name from sometable where name like ${pattern}", pattern='f%d'):
		...		print name
		
		Special keywords:

		'_store' is 0: force saving the result on the server
		'_store' is 1: force saving the result on the client
		Otherwise:     save on the server if the backend supports multiple
		               concurrent cursors on a single connection

		'_head' is 1: first return is headers as text
		'_head' is 2: first return is DB header tuples

		'_dict' is True: yield entries as dictionary instead of list
		'_dict' is a type: as before, but use that. 
		'_empty' is True: don't throw an error when no data are returned
		'_callback': pass rows to a procedure (either as arguments or as
		             keywords, depending on _dict), return row count

		"""
		cb = kv.get('_callback',None)
		if cb:
			n = 0
			for x in self._DoSelect(_cmd, **kv):
				if kv.get('_dict',None):
					cb(**x)
				else:
					cb(*x)
				n += 1
			return n
		else:
			return self._DoSelect(_cmd, **kv)

	def _DoSelect(self, _cmd, **kv):
		conn=self._conn()

		store=kv.get("_store",self.DB._store)

		if self.DB._cursor:
			if store:
				curs=conn.cursor(*self.CArgs)
			elif self.DB.dbtype == "mysql":
				curs=conn.cursor(self.DB.DB.cursors.SSCursor)
			else:
				curs=conn.cursor(*self.CArgs)
		_cmd = self.prep(_cmd, **kv)
		try:
			if self.DB._cursor:
				curs.execute(*_cmd)
			else:
				curs = conn.query(*_cmd)
		except:
			fixup_error(_cmd)
			raise

	
		head = kv.get("_head",None)
		if head:
			if head>1:
				yield curs.description
			else:
				yield map(lambda x:x[0], curs.description)

		as_dict=kv.get("_dict",None)
		if as_dict:
			if as_dict is True:
				as_dict = dict
			names = map(lambda x:x[0], curs.description)

		val = curs.fetchone() if hasattr(curs,'fetchone') else curs.rows.pop(0) if curs.rows else None
		if not val:
			if self._trace is not None:
				self._trace("DoSelect",_cmd,None)

			if '_empty' not in kv:
				raise NoData(_cmd)

		n=0
		while val is not None:
			n += 1
			if as_dict:
				yield as_dict(zip(names,val))
			else:
				yield val[:]
				# need to copy because the array may be re-used
				# internally by the database driver, but the consumer
				# might want to store/modify it
			val = curs.fetchone() if hasattr(curs,'fetchone') else curs.rows.pop(0) if curs.rows else None

		if self._trace is not None:
			self._trace("DoSelect",_cmd,n)


	def __call__(self):
		return self

	def __enter__(self):
		return self

	def __exit__(self, a,b,c):
		if b is None or isinstance(b,CommitThread):
			self.commit()
		else:
			self.rollback()
		return False




#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import generators,absolute_import,print_function

"""\
Simple test script
"""

import os
from sqlmix.async_ import Db,ManyData,NoData
from warnings import filterwarnings

filterwarnings("ignore",category=RuntimeWarning,lineno=15)

async def run_test(x,ai):
 async with Db("db"+str(x),config="test.ini") as dbp:

  try:
        await dbp.Do("drop table if exists test1")
  except NoData:
        pass
  await dbp.Do("""\
        create table test1 (
            id integer primary key not null %s,
            a varchar(255) not null default '',
            b varchar(255) null
        )
    """ % (ai,), _empty=True)

  async with dbp() as db:
    A = await db.Do("insert into test1(a,b) values (${a},${b})", a="one",b="OneOne")
    B = await db.Do("insert into test1(a) values (${a})", a="two")
    C = await db.Do("insert into test1(b) values (${b})", b="three")
  async with dbp() as db:
    async def chk(db):
        a,b = await db.DoFn("select a,b from test1 where id=${id}", id=A)
        assert a=="one"; assert b=="OneOne"
        a,b = await db.DoFn("select a,b from test1 where id=${id}", id=B)
        assert a=="two"; assert b is None
        a,b = await db.DoFn("select a,b from test1 where id=${id}", id=C)
        assert a==""; assert b=="three"
    await chk(db)
    await db.Do("delete from test1 where id=${id}", id=B)
    await db.rollback()

  async with dbp() as db:
    n=0; j=A
    async for i, in db.DoSelect("select id from test1 order by id"):
        n += 1
        assert i==j, (i,j)
        j += 1
    assert n == 3
    try:
        await db.DoFn("select id from test1")
    except ManyData:
        pass
    else:
        assert False,"Need to raise ManyData"

    try:
        async for i, in db.DoSelect("select id from test1 where id < 0"):
            assert False,"Returned nonsense"
    except NoData:
        pass
    else:
        assert False,"Need to raise NoData"

    async for i, in db.DoSelect("select id from test1 where id < 0", _empty=True):
        assert False,"Returned nonsense"
  print("Success.")

async def run_tests():
    await run_test(2,"auto_increment")
#run_test(3,"auto_increment","drop table if exists test1")

import anyio
anyio.run(run_tests, backend="trio")



# -*- coding: utf-8 -*-

from __future__ import generators,absolute_import

"""\
This class is an anyio-compatible frontend to sqlmix.Db.

It has the same interface, except that all Do* methods return a future.
Internally it works by wrapping trio-mysql or aiopg (asyncio backend only).

 >> import sqlmix.async_ as sqlmix
 >> dbi = sqlmix.DbPool([args of sqlmix.Db])
 >>
 >> async def foo(what_id):
 >>     async with dbi() as db:
 >>         await db.Do("delete from whatever where id=${id}", id=what_id, _empty=1)
 >>         async for x in db.DoSelect("select id from whatever", _empty=1):
 >>            print res
 >>         await db.commit()
"""
#
#    Copyright (C) 2011 Matthias urlichs <smurf@smurf.noris.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sqlmix
from sqlmix import NoData,ManyData,fixup_error

from time import time,sleep
import string
import re
import sys
from traceback import print_exc
from contextlib import asynccontextmanager

import anyio

import logging
logger = logging.getLogger(__name__)

__all__ = ('DbPool','NoData','ManyData')

_DEBUG = False

def _call(r,p,a,k):
    """Drop the first argument (i.e. lose the Deferred result)"""
    return p(*a,**k)

def _print_error(f):
    f.printTraceback(file=sys.stderr)

def debug_(flag,*a):
    if not flag: return
    def pr(x):
        if isinstance(x,(tuple,list)):
            return "\n".join((pr(y).strip("\n") for y in x))
        elif isinstance(x,bytes):
            try:
                return unicode(x)
            except:
                return repr(x)
        elif isinstance(x,str):
            return x
        elif isinstance(x,(float,int)):
            return str(x)
        else:
            return repr(x)
    s=" ".join((pr(x) for x in a))

    if flag is True: fl=""
    else: fl=flag+": "
    logger.debug("%s%s",fl,s)

def debug(*a):
    debug_(_DEBUG,*a)

#class db_data(object):
#    sequential = False
#    _store = 1 # safe default
#    _cursor = True
#    def __init__(self, **kwargs):
#        """standard keywords: host,port,database,username,password"""
#        for f in "host port database username password".split():
#            try:
#                v = kwargs[f]
#            except KeyError:
#                pass
#            else:
#                if f == "port":
#                    v=int(v)
#                if v is not None:
#                    setattr(self,f,v)

class _db_mysql(sqlmix.db_data):
    port=3306
    def __init__(self, **kwargs):
        self.DB = __import__("trio_mysql")
        super().__init__(**kwargs)
        self.DB.cursors = __import__("trio_mysql.cursors").cursors
        self.DB.paramstyle = 'format'

    async def _conn(self, task_status):
        with anyio.CancelScope(shield=True) as sc:
            async with self.DB.connect(db=self.database, host=self.host, user=self.username, password=self.password, port=self.port, **self.kwargs) as conn:
                conn._sqlmix_scope = sc
                task_status.started(conn)
                await anyio.sleep_forever()

    async def conn(self, db):
        return await db._tg.start(self._conn)


class _db_postgres(sqlmix.db_data):
    def __init__(self, **kwargs):
        self.DB = __import__("aiopg")
        super().__init__(**kwargs)

    async def conn(self, db):
        res = await self.DB.connect(self.database)
        res._sqlmix_scope = None
        return res

_databases = {
    "mysql": _db_mysql,
    "postgres": _db_postgres,
}

class CtxObj:
    __ctx = None
    async def __aenter__(self):
        if self.__ctx is not None:
            raise RuntimeError("duplicate context")
        self.__ctx = ctx = self._ctx()  # pylint: disable=E1101,W0201
        return await ctx.__aenter__()

    def __aexit__(self, *tb):
        ctx,self.__ctx = self.__ctx,None
        return ctx.__aexit__(*tb)

class Db(CtxObj, sqlmix.DbPrep):
    """\
    Manage a pool of database connections.
    """
    timeout = 70 # one minute plus
    cleaner = None
    _trace = None
    db = None

    def __init__(self,cfg=None, dbtype='mysql', _timeout=None, **kwargs):
        """\
        Create a pool of database connections, for processing (a sequence of)
        SQL commands.
        """

        if cfg is not None:
            try:
                cffile = kwargs.pop('config')
            except KeyError:
                from os.path import expanduser as home
                cffile = home("~/.sqlmix.conf")
            if isinstance(cffile,str):
                from configparser import ConfigParser
                cfp = ConfigParser()
                cfp.read(cffile)
            else:
                cfp = cffile
            args = dict(cfp.items(cfg))
            args.update(kwargs)
            kwargs = args

        if _timeout is not None:
            self.timeout = _timeout

        kwargs.setdefault('use_unicode',True)
        # kwargs.setdefault('no_delay',True)

        self.kwargs = kwargs

        dbtype = kwargs.pop('dbtype','mysql')
        self.DB = _databases[dbtype](**kwargs)
        self.DB.dbtype=dbtype
        if self._trace is not None:
            self._trace("INIT",dbtype,kwargs)

        self.CArgs = ()

        if hasattr(self.DB,'paramstyle'):
            paramstyle = self.DB.paramstyle
        else:
            paramstyle = self.DB.DB.paramstyle
        (self.arg_init, self.arg_do, self.arg_done) \
            = sqlmix._parsers[paramstyle]

        self.db = []
        self.cleaner = None
        self.stopping = False

        super(Db,self).__init__()

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as self._tg:
            try:
                yield self
            finally:
                self.close()


    def stop(self):
        self.stopping = True

    @asynccontextmanager
    async def db(self):
        res = await self._get_db()
        try:
            yield res
        finally:
            self._put_db(res)

    async def _get_db(self):
        if self.db:
            r = self.db.pop()[0]
            s="OLD"
        else:
            r = await self.DB.conn(self)
            s="NEW"


        debug(s)
        return r

    def _put_db(self,db):
        if self.db is None or self.stopping:
            db._sqlmix_scope.cancel()
            return
        t = time()+self.timeout
        self.db.append((db,t))
        debug("BACK",getattr(db,"tid",None))
    
    async def _clean(self):
        self.cleaner = None
        while self.db:
            t = time()
            while self.db and self.db[0][1] <= t:
                db = self.db.pop(0)[0]
                db.close()
            await anyio.sleep(self.db[0][1]-t)

    def close(self):
        self.stopping = True
        if self.cleaner:
            self.cleaner.cancel()
            self.cleaner = None
        while self.db:
            db = self.db.pop(0)[0]
            db._sqlmix_scope.cancel()
        self._tg.cancel_scope.cancel()

    def __call__(self, job=None,retry=0):
        """\
        Get a new connection from the database pool (or start a new one)
        and return a handler object.

        Usage:
        >>> async def something(...):
        >>>     dbpool = Db(...) # arguments like sqlmix.Db()
        >>>     async with dbpool() as db:
        >>>         d = await db.Do("...")
        >>>         do_whatever(d)

        The transaction will be committed if you leave the "with" block
        normally or with a CommitThread exception. Otherwise, it will be
        rolled back.

        Alternately, you can pass a procedure and an optional repeat count:
        >>> def proc(db):
        >>>     d = db.Do("...")
        >>>     return d
        >>> d = await dbpool(proc, 10)

        The procedure will be retried up to 10 times if there are errors;
        if they persist, the first error will be re-raised.

        """
        if job is None:
            if retry:
                raise RuntimeError("You can't use 'retry' without something to call")
            return DbConn(self)
        return self._call(job,retry)

    async def _call(self, job, retry):
        global tid
        tid += 1
        mtid = tid
        err = None
        debug("STARTCALL",job,retry,mtid)

        e1 = None
        try:
            while True:
                db = self._get_db(mtid)
                self._note(db)
                try:
                    debug("CALL JOB",mtid)
                    d = await job(db)
                    debug("RET JOB",mtid,d)
                    def pr(r):
                        debug("RES JOB",mtid,r)
                        return r
                    return d
                except (EnvironmentError,NameError) as e:
                    self._denote(db)
                    await db.rollback()
                    raise
                except Exception as e:
                    self._denote(db)
                    await db.rollback()
                    if e1 is None:
                        e1 = e
                    if retry:
                        retry -= 1
                        continue
                    raise e1
                except BaseException:
                    self._denote(db)
                    await db.rollback()
                    raise
                else:
                    self._denote(db)
                    await db.commit()
                    return res
        finally:
            debug("ENDCALL",job,retry)

    def _note(self,x):
        if not _DEBUG: return
        import inspect
        self._tb[x.tid] = inspect.stack(1)
    def _denote(self,x):
        if not _DEBUG: return
        del self._tb[x.tid]

    async def Do(self,cmd,**kv):
        async with self() as db:
            res = await db.Do(cmd, **kv)
        return res

    async def DoFn(self,cmd,**kv):
        async with self() as db:
            res = await db.DoFn(cmd, **kv)
        return res

    async def DoSelect(self,cmd,**kv):
        raise NotImplementedError("You need to call DoSelect from a transaction")
        
## Py3.6
#   async def DoSelect(self,cmd,**kv):
#       n = 0
#       async with self() as db:
#           async for r in db.DoSelect(cmd,**kv):
#               yield r
#       return n

def _do_callback(tid,d,res):
    debug("DO_CB",tid,d,res)
    d.callback(res)
    debug("DID_CB",tid,d,res)

class DbConn(CtxObj):
    """\
    Manage a single connection.
    """
    curs = None
    db = None

    def __init__(self,pool):
        self.pool = pool
        self.committed = []
        self.rolledback = []
        self._trace = pool._trace

    @asynccontextmanager
    async def _ctx(self):
        assert self.db is None
        assert self.curs is None
        self.db = await self.pool._get_db()
        self.DB = self.pool.DB
        try:
            async with self.db.cursor():
                try:
                    yield self

                except sqlmix.CommitThread:
                    pass
                except Exception as exc:
                    from traceback import format_exception
                    debug("ERROR",format_exception(exc))
                    await self.rollback()
                    raise
                try:
                    await self.commit()
                except sqlmix.CommitThread:
                    pass
                except Exception as exc:
                    await self.rollback()
                    raise
        except Exception:
            self.pool._put_db(self.db)
            raise
        except BaseException:
            self.db._sqlmix_scope.cancel()
            raise
        else:
            self.pool._put_db(self.db)
        finally:
            self.db = None


    def call_committed(self,proc,*a,**k):
        self.committed.append((proc,a,k))
    def call_rolledback(self,proc,*a,**k):
        self.rolledback.append((proc,a,k))

    def _run_committed(self):
        return self._run_(self.committed, "COMMIT")
    def _run_rolledback(self):
        return self._run_(self.rolledback, "ROLLBACK")
    async def _run_(self,rc,rcname):
        self.committed = []
        self.rolledback = []
        for proc,a,k in rc[::-1]:
            debug("AFTER "+rcname,proc,a,k)
            try:
                await _call(proc,a,k)
            except Exception as exc:
                logger.exception(proc)

    def close(self,reason="???"):
        if self.curs is not None:
            self.curs.close()
            self.curs = None
        sc = self.db._sqlmix_scope
        if sc is not None:
            self.db._sqlmix_scope = None
            sc.cancel()

    async def commit(self,res=None):
        await self.db.commit()
        await self._run_committed()

    async def rollback(self,res=None):
        await self.db.rollback()
        await self._run_rolledback()

    async def _cursor(self, cmd, **kv):
        cmd = self.pool.prep(cmd, **kv)
        try:
            async with self.db.cursor() as curs:
                await curs.execute(*cmd)
        except:
            fixup_error(cmd)
            raise
        return curs
        
    async def DoFn(self, cmd, **kv):
        curs = await self._cursor(cmd, **kv)

        if hasattr(curs,'fetchone'):
            val = await curs.fetchone()
        elif not curs.rows:
            val = None
        else:
            val = curs.rows.pop(0)

        if self._trace is not None:
            self._trace("DoFn",cmd,val)
        if not val:
            raise NoData(cmd,kv)

        as_dict=kv.get("_dict",None)
        if as_dict:
            if as_dict is True:
                as_dict = dict
            names = map(lambda x:x[0], curs.description)

        if ((await curs.fetchone()) is not None) if hasattr(curs,'fetchone') else curs.rows:
            raise ManyData(cmd)
        if self.curs is None:
            await curs.aclose()

        if as_dict:
            val = as_dict(zip(names,val))
        return val

    async def Do(self, cmd, **kv):
        """Database-specific Do function"""
        curs = await self._cursor(cmd, **kv)

        r = curs.lastrowid
        if not r:
            r = curs.rowcount
        if self.curs is None:
            await curs.aclose()

        if self._trace is not None:
            self._trace("Do",cmd,r)
        if r == 0 and not '_empty' in kv:
            raise NoData(cmd)
        return r

# Py3.6
#    async def DoSelect(self, cmd, **kv):
#        """Database-specific DoSelect function"""
#        cmd = self.db.prep(cmd, **kv)
#
#        try:
#            curs = await self.cursor()
#
#            await curs.execute(*_cmd)
#        except:
#            fixup_error(_cmd)
#            raise
#
#        n = 0
#        as_dict=kv.get("_dict",None)
#        if as_dict is True:
#            as_dict = dict
#        if as_dict:
#            names = map(lambda x:x[0], curs.description)
#
#        try:
#            while True:
#                if hasattr(curs,'fetchone'):
#                    val = await curs.fetchone()
#                elif not curs.rows:
#                    break
#                else:
#                    val = curs.rows.pop(0)
#
#                if ((await curs.fetchone()) is not None) if hasattr(curs,'fetchone') else curs.rows:
#                    raise ManyData(_cmd)
#                if as_dict:
#                    val = as_dict(zip(names,val))
#
#                yield val
#
#        finally:
#            if self._trace is not None:
#                self._trace("DoSel",_cmd,val)
#            if self.curs is None:
#                curs.close()
#        if n == 0 and not self.maybe_empty:
#            raise NoData(_cmd)

    def DoSelect(selfi,cmd,**kv):
        class SelectCmd(object):
            curs = None
            names = None

            def __init__(self,cmd,**k):
                self.cmd = cmd
                self.k = k

                self.n = 0
                self.as_dict=k.pop("_dict",False)
                self.maybe_empty=k.pop("_empty",False)
                if self.as_dict is True:
                    self.as_dict = dict

            def __aiter__(self):
                return self

            async def __anext__(self):
                if self.curs is None:
                    self.curs = curs = await selfi._cursor(cmd, **kv)

                    if self.as_dict:
                        self.names = list(map(lambda x:x[0], curs.description))
                else:
                    curs = self.curs
                
                if hasattr(curs,'fetchone'):
                    val = await curs.fetchone()
                elif not curs.rows:
                    val = None
                else:
                    val = curs.rows.pop(0)

                if val is None:
                    if self.n == 0 and not self.maybe_empty:
                        raise NoData(self.cmd)
                    raise StopAsyncIteration

                if self.as_dict:
                    val = self.as_dict(zip(self.names,val))

                self.n += 1
                return val

        return SelectCmd(cmd,**kv)

    Do.__doc__ = sqlmix.Db.Do.__doc__ + "\nReturns a Future.\n"
    DoFn.__doc__ = sqlmix.Db.DoFn.__doc__ + "\nReturns a Future.\n"
    DoSelect.__doc__ = sqlmix.Db.DoSelect.__doc__ + "\nReturns a Future.\n"




#!/usr/bin/env python

"""Setup script for smurf's Python database module"""

from distutils.core import setup

description = "Simple syntax for any database"
long_description = \
"""
This module contains database access methods which are very easy
to use and work with any Python database module without touching
parameterized SQL syntax.

"""

setup (name = "sqlmix",
       use_scm_version={"version_scheme": "guess-next-dev", "local_scheme": "dirty-tag"},
       setup_requires=["setuptools_scm"],
       description = description,
       long_description = long_description,
       author = "Matthias Urlichs",
       author_email = "smurf@smurf.noris.de",
        classifiers=[
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Database :: Front-Ends',
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        ],

       url = "https://github.com/smurfix/sqlmix",
       license = 'GPLv3',
       platforms = ['POSIX'],
       packages = ['sqlmix'],
       install_requires=[
           "anyio",
           "trio",
           "trio-mysql",
       ]
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import generators,absolute_import,print_function

"""\
Simple test script
"""

import os
from sqlmix import Db,ManyData,NoData
from warnings import filterwarnings

filterwarnings("ignore",category=RuntimeWarning,lineno=15)

def run_test(x,ai,*a):
	db = Db("db"+str(x),config="test.ini")
	for c in a:
		db.Do(c,_empty=True)
	db.Do("""\
		create table test1 (
			id integer primary key not null %s,
			a varchar(255) not null default '',
			b varchar(255) null
		)
	""" % (ai,), _empty=True)
	db.commit()
	A = db.Do("insert into test1(a,b) values (${a},${b})", a="one",b="OneOne")
	B = db.Do("insert into test1(a) values (${a})", a="two")
	C = db.Do("insert into test1(b) values (${b})", b="three")
	db.commit()
	def chk(db):
		a,b = db.DoFn("select a,b from test1 where id=${id}", id=A)
		assert a=="one"; assert b=="OneOne"
		a,b = db.DoFn("select a,b from test1 where id=${id}", id=B)
		assert a=="two"; assert b is None
		a,b = db.DoFn("select a,b from test1 where id=${id}", id=C)
		assert a==""; assert b=="three"
	chk(db)
	db.Do("delete from test1 where id=${id}", id=B)
	db.rollback()

	n=0; j=A
	for i, in db.DoSelect("select id from test1 order by id"):
		n += 1
		assert i==j
		j += 1
	assert n == 3
	try:
		db.DoFn("select id from test1")
	except ManyData:
		pass
	else:
		assert False,"Need to raise ManyData"

	try:
		for i, in db.DoSelect("select id from test1 where id < 0"):
			assert False,"Returned nonsense"
	except NoData:
		pass
	else:
		assert False,"Need to raise NoData"

	for i, in db.DoSelect("select id from test1 where id < 0", _empty=True):
		assert False,"Returned nonsense"
	print("Success.")

try: os.unlink("test.db")
except EnvironmentError: pass
run_test(1,"")
run_test(2,"auto_increment","create database if not exists test_sqlmix","drop table if exists test1")
run_test(3,"auto_increment","drop table if exists test1")

